@@ -40,12 +40,12 @@ STDMETHODIMP COpenCallbackImp::GetProperty(PROPID propID, PROPVARIANT *value)
   COM_TRY_BEGIN
   NCOM::CPropVariant prop;
   if (_subArchiveMode)
-    switch(propID)
+    switch (propID)
     {
       case kpidName: prop = _subArchiveName; break;
     }
   else
-    switch(propID)
+    switch (propID)
     {
       case kpidName:  prop = _fileInfo.Name; break;
       case kpidIsDir:  prop = _fileInfo.IsDir(); break;
@@ -65,7 +65,9 @@ struct CInFileStreamVol: public CInFileStream
   int FileNameIndex;
   COpenCallbackImp *OpenCallbackImp;
   CMyComPtr<IArchiveOpenCallback> OpenCallbackRef;
-  CInFileStreamVol(bool ignoreLink = false) : CInFileStream(ignoreLink) { } 
+
+  CInFileStreamVol() : CInFileStream(true) { }
+ 
   ~CInFileStreamVol()
   {
     if (OpenCallbackRef)
@@ -73,30 +75,57 @@ struct CInFileStreamVol: public CInFileStream
   }
 };
 
+
+// from ArchiveExtractCallback.cpp
+bool IsSafePath(const UString &path);
+
 STDMETHODIMP COpenCallbackImp::GetStream(const wchar_t *name, IInStream **inStream)
 {
   COM_TRY_BEGIN
   *inStream = NULL;
+  
   if (_subArchiveMode)
     return S_FALSE;
   if (Callback)
   {
     RINOK(Callback->Open_CheckBreak());
   }
+
+  UString name2 = name;
+
+  
+  #ifndef _SFX
+  
+  #ifdef _WIN32
+  name2.Replace(L'/', WCHAR_PATH_SEPARATOR);
+  #endif
+
+  // if (!allowAbsVolPaths)
+  if (!IsSafePath(name2))
+    return S_FALSE;
+  
+  #endif
+
+
   FString fullPath;
-  if (!NFile::NName::GetFullPath(_folderPrefix, us2fs(name), fullPath))
+  if (!NFile::NName::GetFullPath(_folderPrefix, us2fs(name2), fullPath))
     return S_FALSE;
   if (!_fileInfo.Find(fullPath,true))
     return S_FALSE;
   if (_fileInfo.IsDir())
     return S_FALSE;
-  CInFileStreamVol *inFile = new CInFileStreamVol(true);
+  CInFileStreamVol *inFile = new CInFileStreamVol;
   CMyComPtr<IInStream> inStreamTemp = inFile;
   if (!inFile->Open(fullPath))
-    return ::GetLastError();
+  {
+    DWORD lastError = ::GetLastError();
+    if (lastError == 0)
+      return E_FAIL;
+    return HRESULT_FROM_WIN32(lastError);
+  }
 
   FileSizes.Add(_fileInfo.Size);
-  FileNames.Add(name);
+  FileNames.Add(name2);
   inFile->FileNameIndex = FileNames_WasUsed.Add(true);
   inFile->OpenCallbackImp = this;
   inFile->OpenCallbackRef = this;
@@ -119,6 +148,7 @@ STDMETHODIMP COpenCallbackImp::CryptoGetTextPassword(BSTR *password)
   }
   if (!Callback)
     return E_NOTIMPL;
+  PasswordWasAsked = true;
   return Callback->Open_CryptoGetTextPassword(password);
   COM_TRY_END
 }
